import functools
import signal
import sys

import fire
from loguru import logger

from workcraft.models import Workcraft
from workcraft.peon import Peon
from workcraft.utils import import_module_attribute


def handle_shutdown(peon: Peon | None, signum: int, frame) -> None:
    """Handle shutdown signals gracefully"""
    logger.info(f"Received signal {signum}. Initiating graceful shutdown...")
    if peon:
        peon.stop()


def main(workcraft_path: str, queues: list[str] | None = None) -> None:
    peon = None
    try:
        logger.info(f"Getting workcraft object at {workcraft_path}")
        workcraft: Workcraft = import_module_attribute(workcraft_path)

        if queues:
            if isinstance(queues, tuple):
                logger.info(f"Converting queues to list: {queues}")
                queues = list(queues)
            elif isinstance(queues, str):
                logger.info(f"Converting queues to list: {queues}")
                queues = [queues]
            elif not isinstance(queues, list):
                raise ValueError(f"Invalid queues type: {type(queues)}")
            else:
                logger.info(f"Using queues: {queues}, type {type(queues)}")

        peon = Peon(workcraft, queues=queues)
        # Set up signal handlers with peon instance
        signal_handler = functools.partial(handle_shutdown, peon)
        signal.signal(signal.SIGTERM, signal_handler)
        signal.signal(signal.SIGINT, signal_handler)

        # Start work - this will block until stop() is called
        peon.work()

    except Exception as e:
        logger.error(f"Fatal error: {e}")
        if peon:
            peon.stop()
        sys.exit(1)


if __name__ == "__main__":
    fire.Fire(main)



import json
import time

import requests


while True:
    try:
        response = requests.get(
            "http://localhost:6112/events?type=peon&peon_id=abc&queues=['DEFAULT']",
            stream=True,
        )
        for line in response.iter_content(chunk_size=None):
            if line:
                print(line)
                message = line.decode().split("data:")[1]
                message = json.loads(message)
                print(message)
    except requests.exceptions.ConnectionError as e:
        print("failed to connect, likely server offline", e)
        time.sleep(1)
    except Exception as e:
        print("error", e)
        time.sleep(1)



import datetime
import json
import threading
import uuid
from queue import Empty, Queue

import requests
from loguru import logger

from workcraft.models import Task, Workcraft
from workcraft.utils import capture_all_output, tenacious_request


class Peon:
    def __init__(
        self,
        workcraft: Workcraft,
        id: str | None = None,
        queues: list[str] | None = None,
    ) -> None:
        self.id = id or uuid.uuid4().hex
        self.queues = queues
        self.workcraft = workcraft
        self.seen_tasks_in_memory = set()
        self.current_task = None

        self.working = True
        self.queue = Queue()
        self.connected = False
        self._sse_thread = threading.Thread(target=self._run_sse, daemon=True)
        self._heartbeat_thread = threading.Thread(target=self._heartbeat, daemon=True)
        self._processor_thread = threading.Thread(target=self._process, daemon=True)
        self._statistics_thread = threading.Thread(target=self._statistics, daemon=True)

        self._stop_event = threading.Event()
        self._task_cancelled = threading.Event()

    def _sync(self, data: dict) -> None:
        if self.connected:
            try:
                res = tenacious_request(
                    lambda: requests.post(
                        self.workcraft.stronghold_url + f"/api/peon/{self.id}/update",
                        headers={"WORKCRAFT_API_KEY": self.workcraft.api_key},
                        json=data,
                    )
                )

                if 200 <= res.status_code < 300:
                    pass
                else:
                    logger.error(
                        f"Failed to update peon: {res.status_code} - {res.text}"
                    )
            except Exception as e:
                logger.error(f"Failed to send peon update: {e}")

    def work(self) -> None:
        logger.info("Starting peon...")

        self._sse_thread.start()
        logger.info("Started SSE thread")

        self._heartbeat_thread.start()
        logger.info("Started heartbeat thread")
        self._processor_thread.start()
        logger.info("Started processor thread")

        self._statistics_thread.start()
        logger.info("Started statistics thread")
        logger.info(f"Peon ID {self.id}")
        available_tasks = self.workcraft.tasks.keys()
        logger.info("Available Tasks:")
        for task in available_tasks:
            logger.info(f" - {task}")

        logger.success("Zug Zug. Ready to work!")
        try:
            while not self._stop_event.is_set():
                self._stop_event.wait(1)  # Wait with timeout
        except KeyboardInterrupt:
            logger.info("Received keyboard interrupt")
            self.stop()

    def _cancel_task_in_queue(self, task_id: str) -> None:
        # Remove task from queue
        for i in range(self.queue.qsize()):
            task = self.queue.get()
            if task.id != task_id:
                self.queue.put(task)
            else:
                logger.info(f"Removing task {task_id} from queue")
                task.status = "CANCELLED"

                res = tenacious_request(
                    lambda: requests.post(
                        f"{self.workcraft.stronghold_url}/api/task/{task.id}/update",
                        headers={"WORKCRAFT_API_KEY": self.workcraft.api_key},
                        json=Task.to_stronghold(task),
                    )
                )

                if 200 <= res.status_code < 300:
                    logger.info(f"Task {task_id} removed from queue")
                else:
                    logger.error(f"Failed to remove task from queue: {res.text}")

                self.queue.task_done()
                break
        logger.info(f"Task {task_id} removed from queue")

    def _statistics(self) -> None:
        while self.working and not self._stop_event.is_set():
            try:
                res = tenacious_request(
                    lambda: requests.post(
                        self.workcraft.stronghold_url
                        + f"/api/peon/{self.id}/statistics",
                        json={
                            "type": "queue",
                            "value": {
                                "size": self.queue.qsize(),
                            },
                            "peon_id": self.id,
                        },
                        headers={"WORKCRAFT_API_KEY": self.workcraft.api_key},
                    )
                )

                if 200 <= res.status_code < 300:
                    # logger.info("Statistics sent successfully")
                    pass
                else:
                    logger.error(f"Failed to send statistics: {res.text}")
            except Exception as e:
                logger.error(f"Failed to send statistics: {e}")

            self._stop_event.wait(5)

    def _process(self) -> None:
        while self.working and not self._stop_event.is_set():
            try:
                task = self.queue.get(timeout=1)
            except Empty as _:
                self._sync(
                    {
                        "current_task": None,
                        "current_task_set": True,
                        "status": "IDLE",
                        "status_set": True,
                    }
                )
                self.current_task = None
                continue

            try:
                logger.info(f"Processing task {task.id}")
                self._sync(
                    {
                        "current_task": task.id,
                        "current_task_set": True,
                        "status": "WORKING",
                        "status_set": True,
                    }
                )

                res = tenacious_request(
                    lambda: requests.post(
                        f"{self.workcraft.stronghold_url}/api/task/{task.id}/update",
                        headers={"WORKCRAFT_API_KEY": self.workcraft.api_key},
                        json={"status": "RUNNING"},
                    )
                )

                if 200 <= res.status_code < 300:
                    logger.info(f"Task {task.id} set to RUNNING")
                else:
                    raise Exception(f"Failed to set task to RUNNING: {res.text}")

                self._task_cancelled.clear()
                result_queue = Queue()

                def execute_task(_task):
                    try:
                        with capture_all_output(self.workcraft, task.id):
                            result = self.workcraft.execute(_task)
                            result_queue.put(result)
                    except Exception as e:
                        result_queue.put(e)

                task_thread = threading.Thread(target=execute_task, args=(task,))
                task_thread.start()

                # Monitor for cancellation or completion
                cancelled = False
                while task_thread.is_alive():
                    if (
                        self._task_cancelled.is_set() or self._stop_event.is_set()
                    ):  # Check both events
                        logger.info("Task cancellation requested")
                        task_thread.join(timeout=5)
                        if task_thread.is_alive():
                            logger.warning("Task did not stop gracefully")
                        task.status = "CANCELLED"
                        cancelled = True
                        break
                    task_thread.join(timeout=1)

                if not cancelled:
                    try:
                        updated_task = result_queue.get_nowait()
                    except Empty as e:
                        logger.error(
                            f"Failed to get task result: {e} because queue is empty"
                        )
                        task.status = "FAILURE"
                        task.result = (
                            f"Task failed to complete. No result available. Error: {e}"
                        )
                        updated_task = task
                else:
                    updated_task = task  # Use the cancelled task

                try:
                    res = tenacious_request(
                        lambda: requests.post(
                            f"{self.workcraft.stronghold_url}/api/task/{task.id}/update",
                            headers={"WORKCRAFT_API_KEY": self.workcraft.api_key},
                            json=Task.to_stronghold(updated_task),
                        )
                    )

                    if 200 <= res.status_code < 300:
                        logger.info(f"Task updated with status: {updated_task.status}")
                    else:
                        logger.error(
                            f"Failed to update task: {res.status_code} - {res.text}"
                        )
                except Exception as e:
                    logger.error(f"Failed to send task update: {e}")

            except Exception as e:
                logger.error(f"Failed to process task: {e}")
            finally:
                self._sync(
                    {
                        "current_task": None,
                        "current_task_set": True,
                        "status": "IDLE",
                        "status_set": True,
                    }
                )

                self.seen_tasks_in_memory.remove(task.id)
                self.queue.task_done()

            # Break the loop if we're stopping
            if self._stop_event.is_set():
                logger.info("Stopping processor thread")
                break

    def _heartbeat(self) -> None:
        while self.working and not self._stop_event.is_set():
            try:
                self._sync(
                    {
                        "last_heartbeat": datetime.datetime.now().isoformat(),
                        "last_heartbeat_set": True,
                    }
                )
            except Exception as e:
                logger.error(f"Failed to send ping: {e}")
            self._stop_event.wait(5)  # Replace sleep with event wait

    def _queue_to_stronghold(self) -> str:
        if self.queues is None:
            return "[]"
        return "['" + "','".join(self.queues) + "']"

    def _run_sse(self):
        logger.info("Starting SSE thread")

        while self.working and not self._stop_event.is_set():
            print("Connecting to server after 5 seconds...")
            self._stop_event.wait(5)
            try:
                logger.info(f"Attempting connection to {self.workcraft.stronghold_url}")
                response = requests.get(
                    f"{self.workcraft.stronghold_url}/events?type=peon&peon_id={self.id}&queues={self._queue_to_stronghold()}",
                    stream=True,
                    headers={"WORKCRAFT_API_KEY": self.workcraft.api_key},
                )
                if response.status_code != 200:
                    logger.error(f"Failed to connect to server: {response.text}")
                    self._stop_event.wait(5)
                    continue
                buffer = ""
                for line in response.iter_content(chunk_size=None):
                    if line:
                        try:
                            buffer += line.decode()

                            if not buffer.endswith("\n\n"):
                                continue

                            msg = buffer.split("data:")[1]
                            msg = json.loads(msg)

                            buffer = ""
                            logger.info(f"Received message: {msg}")
                            if msg["type"] == "new_task":
                                try:
                                    task = Task.model_validate(msg["data"])
                                except Exception as e:
                                    logger.error(
                                        f"Failed to validate task: {e}, malformed."
                                        " Setting task to INVALID"
                                    )
                                    task_id = msg["payload"]["id"]
                                    if not task_id:
                                        logger.error("Task ID is missing")
                                        continue

                                    res = tenacious_request(
                                        lambda: requests.post(
                                            f"{self.workcraft.stronghold_url}/api/task/{task_id}/update",
                                            headers={
                                                "WORKCRAFT_API_KEY": self.workcraft.api_key
                                            },
                                            json={
                                                "status": "INVALID",
                                                "result": f"Task is invalid: {e}",
                                            },
                                        )
                                    )

                                    if 200 <= res.status_code < 300:
                                        logger.info(f"Task {task_id} set to INVALID")
                                    else:
                                        logger.error(
                                            f"Failed to set task to INVALID: {res.text}"
                                        )
                                    continue

                                if task.id in self.seen_tasks_in_memory:
                                    logger.info(
                                        f"Task {task.id} already seen, skipping"
                                    )
                                    continue

                                try:
                                    res = tenacious_request(
                                        lambda: requests.post(
                                            self.workcraft.stronghold_url
                                            + f"/api/task/{task.id}/update",
                                            headers={
                                                "WORKCRAFT_API_KEY": self.workcraft.api_key
                                            },
                                            json={
                                                "peon_id": self.id,
                                                "status": "ACKNOWLEDGED",
                                            },
                                        )
                                    )

                                    if 200 <= res.status_code < 300:
                                        logger.info(
                                            "Task acknowledgement sent successfully"
                                        )
                                    else:
                                        logger.error(
                                            f"Failed to send task ack: {res.text}"
                                        )

                                    self._sync(
                                        {
                                            "current_task": task.id,
                                            "current_task_set": True,
                                            "status": "PREPARING",
                                            "status_set": True,
                                        }
                                    )
                                    task.peon_id = self.id
                                    self.queue.put(task)
                                    self.seen_tasks_in_memory.add(task.id)

                                except Exception as e:
                                    logger.error(
                                        f"Failed to send task acknowledgement: {e}"
                                    )

                            elif msg["type"] == "cancel_task":
                                task_id = msg["payload"]
                                # either cancel the current task or remove from queue
                                if self.current_task and self.current_task == task_id:
                                    self._task_cancelled.set()
                                    logger.info("Task cancellation acknowledged")
                                else:
                                    self._cancel_task_in_queue(task_id)
                            elif msg["type"] == "connected":
                                self.connected = True
                                logger.info("Connected to server")
                            elif msg["type"] == "heartbeat":
                                logger.info("Received heartbeat. Zug zug.")

                        except IndexError:
                            logger.debug(f"Received non-event line: {line.decode()}")
                            continue
                        except json.JSONDecodeError:
                            logger.warning(f"Received invalid JSON: {line.decode()}")
                            continue
            except requests.exceptions.ConnectionError as e:
                logger.info(
                    f"Failed to retrieve stream, likely because server is offline. "
                    f"Raw error: {e}"
                )
                self._stop_event.wait(5)
            except Exception as e:
                logger.error(f"Failed to receive message: {e}")
                continue
        logger.info("SSE thread stopped")

    def stop(self):
        if not self.working:
            logger.info("Peon already shutting down...")
        else:
            logger.info("Initiating shutdown...")
            self.working = False
            self._stop_event.set()
            self._task_cancelled.set()
            # Set a timeout for joining threads
            timeout = 5
            threads = [
                self._sse_thread,
                self._heartbeat_thread,
                self._processor_thread,
                self._statistics_thread,
            ]

            for thread in threads:
                thread.join(timeout=timeout)
                if thread.is_alive():
                    logger.warning(
                        f"Thread {thread.name} did not terminate within {timeout}s"
                    )

            # clean up the queue and set tasks back to PENDING

            while not self.queue.empty():
                try:
                    task = self.queue.get(timeout=1)
                    task.status = "PENDING"
                    task.peon_id = None

                    res = tenacious_request(
                        lambda: requests.post(
                            f"{self.workcraft.stronghold_url}/api/task/{task.id}/update",
                            headers={"WORKCRAFT_API_KEY": self.workcraft.api_key},
                            json=Task.to_stronghold(task),
                        )
                    )

                    if 200 <= res.status_code < 300:
                        logger.info(f"Task {task.id} reset to PENDING")
                    else:
                        logger.error(f"Failed to reset task {task.id}: {res.text}")

                    self.queue.task_done()

                except Exception as e:
                    logger.error(f"Failed to reset task: {e}")

            logger.info("Stopped peon")



import hashlib
import json
import uuid
from collections.abc import Callable
from datetime import datetime

import requests
from beartype.typing import Any, Literal, Protocol
from loguru import logger
from pydantic import BaseModel, Field


class WebsocketMessage(BaseModel):
    type: str
    payload: Any


class TaskPayload(BaseModel):
    task_args: list = Field(default_factory=list)
    task_kwargs: dict = Field(default_factory=dict)
    prerun_handler_args: list = Field(default_factory=list)
    prerun_handler_kwargs: dict = Field(default_factory=dict)
    postrun_handler_args: list = Field(default_factory=list)
    postrun_handler_kwargs: dict = Field(default_factory=dict)


class Task(BaseModel):
    id: str
    task_name: str
    status: Literal[
        "PENDING", "RUNNING", "SUCCESS", "FAILURE", "INVALID", "CANCELLED"
    ] = Field(default="PENDING")
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)
    peon_id: str | None = None
    queue: str = Field(default="DEFAULT")
    payload: TaskPayload
    result: Any | None = None
    retry_on_failure: bool = Field(default=False)
    retry_count: int = Field(default=0)
    retry_limit: int = Field(default=3)

    @classmethod
    def to_stronghold(cls, task: "Task") -> dict:
        """Converts a Task instance to a dictionary for Stronghold."""
        return {
            "id": task.id,
            "task_name": task.task_name,
            "status": task.status,
            "created_at": task.created_at.isoformat() + "Z",
            "updated_at": task.updated_at.isoformat() + "Z",
            "peon_id": task.peon_id,
            "queue": task.queue,
            "payload": task.payload.model_dump_json(),
            "result": json.dumps(task.result),
            "retry_on_failure": task.retry_on_failure,
            "retry_count": task.retry_count,
            "retry_limit": task.retry_limit,
        }


class SetupHandlerFn(Protocol):
    def __call__(self): ...


class TaskHandlerFn(Protocol):
    def __call__(
        self,
        task_id: str,
        *args,
        **kwargs,
    ) -> Any: ...


class PostRunHandlerFn(Protocol):
    def __call__(
        self,
        task_id: str,
        task_name: str,
        result: Any,
        status: Literal["FAILURE", "SUCCESS", "RUNNING", "PENDING"],
        *args,
        **kwargs,
    ): ...


class PreRunHandlerFn(Protocol):
    def __call__(
        self,
        task_id: str,
        task_name: str,
        *args,
        **kwargs,
    ): ...


class Workcraft:
    """Workcraft: A simple distributed task system."""

    def __init__(self, stronghold_url: str, api_key: str):
        self.stronghold_url = stronghold_url

        m = hashlib.sha256()
        m.update(api_key.encode())
        self.api_key = m.hexdigest()

        self.tasks: dict[str, Callable] = {}
        self.setup_handler_fn: Callable | None = None
        self.prerun_handler_fn: Callable | None = None
        self.postrun_handler_fn: Callable | None = None

    def task(self, name: str):
        def decorator(func: TaskHandlerFn):
            self.tasks[name] = func
            return func

        return decorator

    def prerun_handler(self):
        def decorator(func: PreRunHandlerFn):
            self.prerun_handler_fn = func
            return func

        return decorator

    def postrun_handler(self):
        def decorator(func: PostRunHandlerFn):
            self.postrun_handler_fn = func
            return func

        return decorator

    def setup_handler(self):
        def decorator(func: SetupHandlerFn):
            self.setup_handler_fn = func
            return func

        return decorator

    def execute(self, task: Task) -> Task:
        try:
            if self.prerun_handler_fn:
                try:
                    self.prerun_handler_fn(
                        task.id,
                        task.task_name,
                        *task.payload.prerun_handler_args,
                        **task.payload.prerun_handler_kwargs,
                    )
                except Exception as e:
                    logger.warning(f"Failed to execute prerun handler: {e}")
                    logger.warning(
                        f"Task {task.id} will still be executed, "
                        "but the prerun handler failed. This may cause issues "
                        "with the task execution. Please check the logs."
                    )

            task_fn = self.tasks[task.task_name]
            result = task_fn(
                task.id, *task.payload.task_args, **task.payload.task_kwargs
            )
            task.result = json.dumps(result)
            task.status = "SUCCESS"

            if self.postrun_handler_fn:
                try:
                    self.postrun_handler_fn(
                        task.id,
                        task.task_name,
                        result,
                        task.status,
                        *task.payload.postrun_handler_args,
                        **task.payload.postrun_handler_kwargs,
                    )
                except Exception as e:
                    logger.warning(f"Failed to execute postrun handler: {e}")
                    logger.warning(
                        f"Task {task.id} was executed successfully, "
                        "but the postrun handler failed. Please check the logs."
                    )

        except Exception as e:
            import traceback

            print(traceback.format_exc())
            logger.error(f"Failed to execute task: {e}")
            task.status = "FAILURE"
            task.result = str(e)
        finally:
            task.updated_at = datetime.now()
            return task

    def send_task_sync(
        self,
        payload: TaskPayload,
        task_name: str,
        queue: str = "DEFAULT",
        retry_on_failure: bool = False,
        retry_limit: int = 3,
    ) -> str:
        id = str(uuid.uuid4())
        try:
            task = {
                "id": id,
                "task_name": task_name,
                "payload": payload.model_dump(),
                "queue": queue,
                "retry_on_failure": retry_on_failure,
                "retry_limit": retry_limit,
            }
            requests.post(
                f"{self.stronghold_url}/api/task",
                json=task,
                headers={
                    "Content-Type": "application/json",
                    "WORKCRAFT_API_KEY": self.api_key,
                },
            )
        except Exception as e:
            logger.error(f"Failed to send task: {e}")
            raise e
        return id


class State(BaseModel):
    id: str
    status: Literal["IDLE", "PREPARING", "WORKING", "OFFLINE", "PREPARING"]
    current_task: str | None = None
    queues: list[str] | None = None
    last_heartbeat: datetime = Field(default_factory=datetime.now)

    def update_and_return(self, **kwargs) -> "State":
        for key, value in kwargs.items():
            setattr(self, key, value)
        setattr(self, "last_heartbeat", datetime.now())
        return self

    def queue_to_stronghold(self) -> str:
        if self.queues is None:
            return "[]"
        return "['" + "','".join(self.queues) + "']"

    def to_stronghold(self) -> dict:
        queues = None
        if self.queues is not None:
            queues = "['" + "','".join(self.queues) + "']"

        update_json = {
            "current_task": self.current_task if self.current_task else None,
            "current_task_set": bool(self.current_task),
            "status": self.status,
            "status_set": True,
            "last_heartbeat": self.last_heartbeat.isoformat(),
            "last_heartbeat_set": True,
            "queues": queues,
            "queues_set": bool(self.queues),
            "id": self.id,
        }
        return update_json



import asyncio
import logging
import sys
import time

from loguru import logger

from workcraft.models import TaskPayload, Workcraft


stronghold_url = "http://localhost:6112"
api_key = "abcd"
workcraft = Workcraft(stronghold_url, api_key)

global_counter = 0


# @workcraft.setup_handler()
# def setup_handler():
#     global global_counter
#     global_counter = 1000
#     logger.info("Setting up the worker!")


@workcraft.task("simple_task")
def simple_task(task_id: str, a: str) -> int:
    print(task_id, len(a))
    print("Regular print")
    logging.info("Standard logging")
    logger.info("Loguru logging")
    sys.stderr.write("Direct stderr write\n")
    time.sleep(10)
    # raise ValueError("This is a test error mon")

    return 0


@workcraft.prerun_handler()
def prerun_handler(task_id, task_name):
    logger.info(f"PR called for {task_id} and {task_name}!")


@workcraft.postrun_handler()
def postrun_handler(task_id, task_name, result, status):
    logger.info(
        f"PR called for {task_id} and {task_name}! Got {result} and status {status}"
    )


async def main():
    n_tasks = 1
    for _ in range(n_tasks):
        workcraft.send_task_sync(
            task_name="simple_task",
            payload=TaskPayload(
                task_args=["aaaaaa"],
            ),
            retry_on_failure=True,
        )


if __name__ == "__main__":
    asyncio.run(main())



import importlib
import io
import logging
import sys
from collections.abc import Callable
from contextlib import contextmanager

import requests
from loguru import logger
from tenacity import (
    retry,
    retry_if_not_result,
    stop_after_attempt,
    wait_exponential,
)

from workcraft.models import Workcraft


def import_module_attribute(path: str):
    module_path, attr_name = path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, attr_name)


def is_request_successful(response: requests.Response):
    return 200 <= response.status_code < 300


@retry(
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    retry=retry_if_not_result(is_request_successful),
)
def tenacious_request(req_fn: Callable[..., requests.Response]):
    return req_fn()


class TeeIO:
    """Captures output AND forwards it to original stdout/stderr"""

    def __init__(self, original_stream, string_buffer):
        self.original_stream = original_stream
        self.string_buffer = string_buffer

    def write(self, data):
        self.original_stream.write(data)
        self.string_buffer.write(data)

    def flush(self):
        self.original_stream.flush()
        self.string_buffer.flush()


@contextmanager
def capture_all_output(workcraft: Workcraft, task_id: str):
    """Captures stdout, stderr, logging, and loguru output"""
    string_buffer = io.StringIO()

    old_stdout = sys.stdout
    old_stderr = sys.stderr

    sys.stdout = TeeIO(old_stdout, string_buffer)
    sys.stderr = TeeIO(old_stderr, string_buffer)

    log_handler = logging.StreamHandler(string_buffer)
    root_logger = logging.getLogger()
    root_logger.addHandler(log_handler)

    loguru_handler_id = logger.add(string_buffer, format="{message}")

    try:
        yield string_buffer
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        root_logger.removeHandler(log_handler)
        logger.remove(loguru_handler_id)

        logs = string_buffer.getvalue()

        req = requests.post(
            f"{workcraft.stronghold_url}/api/task/{task_id}/update",
            headers={
                "WORKCRAFT_API_KEY": workcraft.api_key,
                "accept": "application/json",
            },
            json={"logs": logs, "logs_set": True},
        )

        if req.status_code != 200:
            logger.error(f"Failed to update logs: {req.text}")

        string_buffer.close()
